"""
Author: Filip Lux (2023), Masaryk University
Licensed under MIT License
"""

import os
import math

import pandas as pd
import numpy as np
from tqdm import tqdm


from pathlib import Path
from .sys_tools import run_procedure
from .libct import run_libct, libct2ctc
from .graph import FlowGraph
from .divergence_tools import compute_divergence_v2


from itertools import repeat, chain
from matplotlib import pyplot as plt

class GlobalTracker():
    
    '''
    one instance per a sequence
    operates in a one directory XX_DATA, where outputs of embedtrack are stored
    '''
    
    def __init__(self,
                 data_path,
                 res_path,
                 lbd=.5,
                 mean_dist=10,
                 max_dist=20,
                 vertex_thr=0.95,
                 app_cost=10000,
                 edge_min_cost=-1):
        '''
        Parameters:
        -----------
        data_path : str
            path to the raw data
            schema of the directory:
            
            <data_path>/01
                       /01_GT
                       
                       # added directories
                       /01_RES                 # symlink to <res_path>/01_RES
                       /01_DATA/mask0000.tif   # output of the segmentation procedure
                               /...
                               /tra_offsets.csv
                               /...
                       ...
        res_path : str
            path to the stored results
            
            <res_path>/01_RES/tracking.txt      # self.run_tracking()
                             /libct.sol
                             /...
                             /mask0000.tif      # self.libctsol2ctc()          
                             /...
                             /res_track.txt     # self.evaluate_ctc()
                             /...
        '''
        
        sequence = os.path.basename(res_path)[:2]
        assert sequence == os.path.basename(data_path)[:2]
            
        self.res_path = res_path
        self.data_path = data_path
        assert os.path.isdir(data_path), data_path
        assert os.path.isdir(res_path), res_path
        
        self.graph = None
        self.sequence = sequence
        self.lbd=lbd
        self.mean_dist = mean_dist
        self.max_dist = max_dist
        self.vertex_thr = vertex_thr
        self.app_cost = app_cost
        self.edge_min_cost = edge_min_cost

    def run_tracking(self,
                     idx_min=0,
                     idx_max=10000,
                     limit_dist=100,
                     n_neighbours=2):

        # compute divergence
        div_path = os.path.join(self.data_path, 'edge_prob_divergence.csv')
        if not os.path.isfile(div_path):
            print('computing divergence')
            compute_divergence_v2(self.data_path)
            
        assert os.path.isfile(div_path)

        # create graph
        print('creating graph')
        self.graph = get_graph_embedtrack(
                        self.data_path,
                        min_frame=idx_min,
                        max_frame=idx_max,
                        limit_dist=limit_dist,
                        n_neighbours=n_neighbours,
                        mean_dist=self.mean_dist,
                        max_dist=self.max_dist)

        # create and save tracking file
        print('creating tracking file')
        tracking = graph2tracking(self.graph,
                                  lbd=self.lbd,
                                  vertex_thr=self.vertex_thr,
                                  max_cost=self.app_cost,
                                  mean_dist=self.mean_dist,
                                  max_dist=self.max_dist,
                                  edge_min_cost=self.edge_min_cost)
        save_tracking(tracking, self.res_path)

        # compute solution
        print('computiong solution')
        run_libct(self.res_path) 
        
    
    def save_results_ctc(self):
        
        # translate libct.sol to ctc result format
        lm = libct2ctc(data_path=self.data_path,
                       res_path=self.res_path)
        return lm  
    
    def solution_stats(self, plot=False):
        '''
        outputs 'solution_stats.txt' file, describing
        a contribution of individual items on a solution
        '''
        
        tracking_file_path = Path(self.res_path, 'tracking.txt')
        sol_file_path = Path(self.res_path, 'tracking.sol')
        assert os.path.isfile(tracking_file_path), tracking_file_path
        assert os.path.isfile(sol_file_path), sol_file_path
        
        sol_stats(sol_file_path, tracking_file_path, plot=plot)
        
        
    def evaluate_ctc(self):
        '''
        computes CTC metrics to evaluate the results
        '''
        
        # TODO: test if the data_path containt GT directory
        
        dirname = os.path.dirname(self.data_path)
        directory = os.path.basename(self.res_path)
        
        gt_path = os.path.join(dirname, directory.replace('RES', 'GT'))
        if not os.path.isdir(gt_path):
            print(f'ground truth masks are not available {gt_path}')
            return
        
        temp_res_path = os.path.join(os.getcwd(), dirname, directory)
        src_path = os.path.join(os.getcwd(), self.res_path)
        
        print(src_path)
        
        # try to remove symlink
        run_procedure(f'rm {temp_res_path}')
        
        assert not Path(temp_res_path).exists(), f'the target folder already exists {temp_res_path}'
        assert Path(src_path).exists()
        
        # create symlink from results to original data folder
        run_procedure(f'ln -s {src_path} {temp_res_path}')
        
        # run evaluation procedures
        measures = ['TRAMeasure', 'DETMeasure', 'SEGMeasure']
        #measures = ['TRAMeasure', 'SEGMeasure']
        for m in measures:
            command = f'./ctc_metrics/{m} {dirname} {self.sequence} 4'
            print(f'running {command}')
            run_procedure(command)
            
        # remove symlink
        run_procedure(f'rm {temp_res_path}')


def compute_edge_cost(log_kl_divergence, distance, avg_dist=10):

    # A) cost depends only on a distance of objects
    cost = distance - avg_dist


    # B) combine time 5
    #cost = cost + np.log1p(np.maximum((dist - avg_dist)**3, 0))

    #cost = cost - (((dist - avg_dist) / 9) **3)

    return - cost


def get_graph_embedtrack(
        res_path,
        min_frame=0,
        max_frame=2000,
        virtual_edge_prob=50,  # equal to divergence_thr=50 in divergence_tools.py
        limit_dist=50,
        n_neighbours=3,
        mean_dist=10,
        max_dist=20,
        edge_min_cost=-1
    ):
    '''
    creates instance of candidate graph
    
    Parameters
    ----------
    res_path: str
        path to the dataset
    min_frame: int
        first frame to analyze 
    max_frame: int 
        last frame to analyze
        
    Returns
    -------
    
    graph: tuple
        structured data about the graph
    
    '''
    
    # create consistent graph structure
    graph = FlowGraph(res_path)

    #pd_edge_path = os.path.join(res_path, 'edge_prob.csv')
    pd_edge_path = os.path.join(res_path, 'edge_prob_divergence.csv')
    pd_vertex_path = os.path.join(res_path, 'vertex_prob.csv')
    assert os.path.isfile(pd_edge_path), pd_edge_path
    assert os.path.isfile(pd_vertex_path), pd_vertex_path

    df_edges = pd.read_csv(pd_edge_path, index_col=False)
    df_vertices = pd.read_csv(pd_vertex_path, index_col=False)
    
    # compute cost
    df_edges['cost'] = np.log(df_edges.divergence)
        
    # list of real frame indexes
    frame_indexes = df_vertices['time'].unique()
    
    # TODO: sort in reverse order
    frame_indexes.sort()
    min_frame = int(np.maximum(frame_indexes.min(), min_frame))
    max_frame = int(np.minimum(frame_indexes.max(), max_frame))
    
    ########## 
    # ADD ALL VERTICES
    print('GET GRAPH: adding vertices')
    for i, row in df_vertices.iterrows():

        prob = row['prob']
        label = row['label']
        frame_idx = row['time']
        
        if not (min_frame <= frame_idx <= max_frame):
            continue
        
        assert label > 0
        
        graph.add_vertex(frame_idx, label, prob)
        
        ########## 
        # ADD SOURCE AND SINK EDGES
        if int(frame_idx) == min_frame:
            graph.add_source_edge(frame_idx, label)
            
        elif int(frame_idx) == max_frame:
            graph.add_sink_edge(frame_idx, label)
        
    ###########
    # ADD ALL MIDDLE EDGES
    print('GET GRAPH: adding edges')
    for i, row in df_edges.iterrows():
        
        #prob = row['prob']
        cost = row['cost']
        label_curr = row['label_curr']
        frame_curr = row['time_curr']
        label_prev = row['label_prev']
        frame_prev = row['time_prev']
        
        if (frame_prev < min_frame) or (frame_curr > max_frame):
            continue
        
        # no appearance in the sequence
        graph.add_edge(frame_curr, label_curr, frame_prev, label_prev, cost)
        
        
    ###########
    # ADD VIRTUAL edges
    # for every edge, that has no connection to t+1:
    #     - add adges to three closest edges (up to 'limit_dist')
    #     - the edge prob is 'virtual_edge_prob' (=.25)
    # TODO: replace with creating full graph only based on a distance
    #       virtual edges then only supports predicted 
    
    print(f'GET GRAPH: entangling to {n_neighbours} neighbours')
    if n_neighbours > 0:
        
        # to_me, from_me = [], []   # REMOVE
        for v_idx in graph.vertex_map.keys():

            if len(graph.edge_to_me[v_idx]) < n_neighbours:
                from_me_ = [graph.edges[e_idx][1] for e_idx in graph.edge_from_me[v_idx]]

                frame_idx, label, x2, y2 = graph.vertex_map[v_idx]
                if frame_idx == min_frame:
                    continue

                # to_me.append(v_idx)  REMOVE
                neighbors = graph.knn(frame_idx-1, x2, y2,
                                      n=n_neighbours,
                                      limit_dist=limit_dist)  # returns list of v_idxs

                # add virtual edges
                for n_v_idx in neighbors:
                    
                    # skip already included neighbours
                    if n_v_idx in from_me_:
                        continue

                    frame_prev, label_prev, x1, y1 = graph.vertex_map[n_v_idx]

                    graph.add_edge(frame_idx,
                                   label,
                                   frame_prev,
                                   label_prev,
                                   virtual_edge_prob)

                    ''' REMOVE
                    cost = get_distance_cost(x1, y1, x2, y2,
                                             mean_dist=mean_dist,
                                             max_dist=max_dist,
                                             min_cost=edge_min_cost)
                    graph.add_edge(frame_idx, label, frame_prev, label_prev, cost)
                    '''

    ''' OUTPUT '''
    indexes = graph.vertex_index, graph.edge_index
    probs = graph.edge_prob, graph.vertex_prob
    move_edges = graph.edge_to_me, graph.edge_from_me

    return indexes, probs, graph.edges, move_edges, graph.vertex_map


def get_distance_cost(dist,
                      mean_dist=10,
                      max_dist=20,
                      min_cost=-1):
    #dist = np.linalg.norm(np.array([x2 - x1, y2 - y1]))
    #return -max((dist - max_dist) / (max_dist - mean_dist), min_cost)
    # TODO: include mean_dist
    return - (dist * min_cost / max_dist - min_cost)
    #return ((dist/max_dist)**2 - 1) * (-min_cost)


# GRAPH TO TRACKING
# KL
def vertex_cost(prob,
                vertex_thr=0.95,
                min_cost=5):
    '''
    probs of real vertices is higher than .95
    '''
    # TODO: extravagant hand-made metrics
    #       replace 
    
    return - min_cost / (1 - vertex_thr) * ( prob - vertex_thr)


# DEPRECIATED
def edge_cost(prob, x1, y1, x2, y2, lbd=.5, mean_dist=10, max_dist=20, min_cost=-1):
    '''
    edge cost is log(divergence)
    '''

    dist = np.linalg.norm(np.array([x2 - x1, y2 - y1]))
    dst_cost = - get_distance_cost(dist, mean_dist, max_dist, min_cost)
    dvg_cost = - prob
        
    return lbd * dst_cost + (1 - lbd) * dvg_cost


def graph2tracking(graph,
                   max_cost=10000,
                   lbd=.5,
                   vertex_thr=0.95,
                   mean_dist=10,
                   max_dist=20,
                   edge_min_cost=-1):
    
    # decompose graph
    indexes, probs, edges, move_edges, vertex_map = graph

    vertex_index, edge_index = indexes
    edge_prob, vertex_prob = probs
    edge_to, edge_from = move_edges

    # variables
    tracking = [] 

    last_frame_index = np.max([f_idx for f_idx, _, _, _ in vertex_map.values()])
    first_frame_index = np.min([f_idx for f_idx, _, _, _ in vertex_map.values()])
    
    frame_shift = first_frame_index
    
    # vertex hypothesis
    # iterate over vertex_map
    
    for v_idx in tqdm(vertex_map.keys()):
        
        frame_idx, label, x, y = vertex_map[v_idx]
        prob = vertex_prob[v_idx]
        
        # compute cost
        cost = vertex_cost(prob, vertex_thr=vertex_thr)
        
        #unique_id = f'V_{time}_{label}' 
        unique_id = f'1{frame_idx:05d}{label:05d}' 
        
        assert len(unique_id) == 11, unique_id
        
        tracking.append(f'H {frame_idx-frame_shift} {unique_id} {cost} {x} {y}')
        
        # no appearance and diappearance -> cost 10000
        app_cost, disapp_cost = max_cost, max_cost
        if (frame_idx == first_frame_index):
            app_cost = 0.
        if (frame_idx == last_frame_index):
            disapp_cost = 0.

        # add appearance and disappearance
        unique_id_app = f'4{frame_idx:05d}{label:05d}' 
        unique_id_dis = f'5{frame_idx:05d}{label:05d}' 
        tracking.append(f'APP {unique_id_app} {unique_id} {app_cost}')
        tracking.append(f'DISAPP {unique_id_dis} {unique_id} {disapp_cost}')

    edges_ = {}
    for e_idx in edges.keys():
        
        v_idx_curr, v_idx_prev  = edges[e_idx]
        
        # appearance
        if v_idx_prev in [0, 1]:
            continue
            
        if v_idx_curr == 0:
            continue
        
        frame_idx_prev, label_prev, x1, y1 = vertex_map[v_idx_prev]
        frame_idx_curr, label_curr, x2, y2 = vertex_map[v_idx_curr]

        dist = np.linalg.norm(np.array([x2 - x1, y2 - y1]))

        # embed cost - bonus term, only negative
        embed_cost = min(edge_prob[e_idx], 0)
        embed_cost = edge_prob[e_idx]

        # dist cost - standard term, zero for ~.95 percentile
        dist_cost = get_distance_cost(dist,
                                      mean_dist=mean_dist,
                                      max_dist=max_dist,
                                      min_cost=edge_min_cost)

        # final cost, weighted by lambda
        cost = lbd * dist_cost + (1-lbd) * embed_cost

        # create the edge descriptor
        seg_id_right = f'1{frame_idx_curr:05d}{label_curr:05d}' 
        seg_id_left = f'1{frame_idx_prev:05d}{label_prev:05d}' 
        unique_id = f'2{seg_id_left}{seg_id_right}'
        assert len(unique_id) == 23, unique_id
        
        # update tracking
        tracking.append(f'MOVE {unique_id} {seg_id_left} {seg_id_right} {cost}')

        # propagate info to compute division events later on
        edges_[seg_id_left] = edges_.get(seg_id_left, [])
        edges_[seg_id_left].append((seg_id_right, cost, dist_cost))

    # division
    div_idx = 0
    for id_left in tqdm(edges_.keys(), 'divisions'):
        
        right_ids = edges_[id_left]
                
        for id_right1, cost1, dist_cost1 in right_ids:
            for id_right2, cost2, dist_cost2 in right_ids:
                if id_right1 >= id_right2:
                    continue
                    
                # define unique id
                unique_id = f'3{id_left}{id_right1}{id_right2}'
                assert len(unique_id) == 34, unique_id
                
                # experimental cost
                #cost = - max(cost1, cost2)
                
                # experimental cost
                # penalize if costs are different
                #diff_penalty = (dist_cost1 - dist_cost2)**2
                #cost = (cost1 + cost2) // 2 + diff_penalty / 10

                cost = (cost1 + cost2) // 2
                
                if cost > 1:
                    continue
                
                tracking.append(f'DIV {unique_id} {id_left} {id_right1} {id_right2} {cost}')
                                
                div_idx += 1
                
    print(f'Added {div_idx} division candidates.')
    return tracking


def save_tracking(tracking, data_path):
    
    tracking_path = os.path.join(data_path, 'tracking.txt')
    with open(tracking_path, 'w', encoding='utf8') as f:
        line_end = repeat("\n")
        lines = chain.from_iterable(zip(tracking, line_end))
        f.writelines(lines)
        
        
        
def sol_stats(sol_path,
              tra_path,
              plot=False):
    '''
    Displays candidate graph 

    Parameters
    ----------

    sol_path : str
        path to 'tracking.sol' file, in libct format
    tra_path : str
        path to 'tracking.txt' file, in libct format
    plot : bool
        plot the histograms

    '''
    

    assert os.path.isfile(tra_path), tra_path
    assert os.path.isfile(sol_path), sol_path
    dirname = os.path.dirname(tra_path)
    
    # COLORS 
    
    # get vertex map from 'tracking.txt'
    vertex_map = {} # vertex_id : (x, y, frame_idx)
    
    tra = {'H': {},
           'APP': {},
           'DISAPP': {},
           'MOVE': {},
           'DIV': {}}
    
    all_tra = {'H': [],
               'APP': [],
               'DISAPP': [],
               'MOVE': [],
               'DIV': []}
        
    with open(tra_path, 'r') as f:
        

        
        # show vertices
        for line in tqdm(f.readlines()):
            tokens = line.split(' ')
            key = tokens[0]
            if key == 'H':
                unique_id, cost = tokens[2], tokens[3]
            elif key == 'APP':
                unique_id, cost = tokens[1], tokens[3]
            elif key == 'DISAPP':
                unique_id, cost = tokens[1], tokens[3]
            elif key == 'MOVE':
                unique_id, cost = tokens[1], tokens[4]
            elif key == 'DIV':
                unique_id, cost = tokens[1], tokens[5]
            else:
                assert False, key
                
            tra[key][unique_id] = cost
            all_tra[key].append(cost)
            
            
    sol = {'H': [],
           'APP': [],
           'DISAPP': [],
           'MOVE': [],
           'DIV': []}
                
    with open(sol_path, 'r') as f:
        
        # show vertices
        for line in tqdm(f.readlines()):
            
            key, unique_id = line.rstrip().split(' ')
            cost = tra[key][unique_id]
            
            sol[key].append(cost)
            
    dir_name = os.path.dirname(sol_path)
    res_path = os.path.join(dir_name, 'sol_stats.txt')
    with open(res_path, 'w') as f:
        
        lines = []
        
        # show vertices
        for key in sol.keys():
            
            # tra
            vals = np.array(all_tra[key], dtype=float)
            line = f'TRA {str(key)} {vals.sum()} {len(vals)} {vals.min()} {vals.mean()} {vals.max()}\n'
            
            lines.append(line)
            
            # sol
            vals = np.array(sol[key], dtype=float)
            if len(vals) > 0:
                line = f'SOL {str(key)} {vals.sum()} {len(vals)} {vals.min()} {vals.mean()} {vals.max()}\n'

                lines.append(line)
        f.writelines(lines)
        
    if plot:
        plt.figure()
        
        for key in sol.keys():
            
            vals = np.array(sol[key], dtype=float)
            plt.hist(vals,
                     histtype='step',
                     density=True,
                     bins=100)
        plt.legend(sol.keys())
        plt.title('tracking.sol')
        plt.xlim((-5, 5))
        plt.ylim((0, 5))
        plt.savefig(os.path.join(dirname, 'sol.png'))
        plt.show()
        
        
        plt.figure()
        
        for key in sol.keys():
            
            vals = np.array(all_tra[key], dtype=float)
            plt.hist(vals,
                     histtype='step',
                     density=True,
                     bins=100)
        plt.legend(sol.keys())
        plt.title('tracking.txt')
        plt.xlim((-5, 5))
        plt.ylim((0, 5))
        plt.savefig(os.path.join(dirname, 'tra.png'))
        plt.show()

        

        
            
        